;; Generated package description from context-coloring.el  -*- no-byte-compile: t -*-
(define-package "context-coloring" "8.1.0" "Highlight by scope" '((emacs "24.3")) :commit "1b30a28857727202d1f6a027f83ad66151fb1e92" :authors '(("Jackson Ray Hamilton" . "jackson@jacksonrayhamilton.com")) :maintainer '("Jackson Ray Hamilton" . "jackson@jacksonrayhamilton.com") :keywords '("convenience" "faces" "tools") :url "https://github.com/jacksonrayhamilton/context-coloring")
